#include "hmbdc/app/Base.hpp"

#include <boost/format.hpp>

#include <iostream>
#include <iomanip>
#include <memory>
#include <unistd.h>
#include <unordered_set>

namespace hmbdc { namespace app { namespace utils {

namespace consoleclient_detail {
using namespace std;
/**
 * @class ConsoleClient<>
 * @brief a Client that work as a console to send and receive network messages
 * @details a Client that receives all net messages delivered to the process
 *  it uses JustBytes to indicate it wants all the messages bytes
 * regardless of their types - still Topic filtering is effective
 * 
 * @tparam NetCtx some NetContext type for a specific transport mechanism, 
 * for example mcast::NetContext, tcpcast::NetContext
 */
template <typename NetCtx>
struct ConsoleClient 
: Client<ConsoleClient<NetCtx>, JustBytes> {
    /**
     * @brief constructor
     * @details just construct the Client - need to be run in the Context later
     * 
     * @param ctx a runtime sized context that the Client will run in
     * @param myCin for incoming commands 
     * @param myCout for output
     * @param myCerr for error
     */
    ConsoleClient(Context<>& ctx
        , istream& myCin = cin
        , ostream& myCout = cout
        , ostream& myCerr = cerr
        , string initCmd = ""
        , std::unordered_set<uint16_t> memoryAttachmentTags = std::unordered_set<uint16_t>()
    ) : ctx_(ctx)
    , hex_(true)
    , myCin_(myCin)
    , myCout_(myCout)
    , myCerr_(myCerr)
    , sender_(nullptr) 
    , initCmd_(initCmd)
    , memoryAttachmentTags_(memoryAttachmentTags) {
        thread stdinThread([this]() {
            stdinThreadEntrance();
        });
        stdinThread_ = move(stdinThread);
    }

    char const* hmbdcName() const { 
        return "console"; 
    }

    /**
     * @brief documentation for all commands the console interprets
     * @details commands are seperated by \n, when read eof, terminates console
     * @return a documentation string 
     */
    static char const* help() {
        return 
"\nCONSOLE LANGUAGE:\n"
"topic <topic>, set current topic\nexample: topic console\n\n"
"ohex, output message in hex format\nexample: ohex\n\n"
"sendhex <tag> <hex bytes>, send a binary message with a message tag\nexample: sendhex 1001 01 00 0f ef\n\n"
"ostr, output message in string format\nexample: ostr\n\n"
"sendstr <tag> <string>, send a string message with a message tag\nexample: sendstr 1001 hello\n\n"
"listen <topic>, console start to listen to the topic\nexample: listen my_topic\n\n"
"stoplisten <topic>, console stop listen to the topic\nexample: stoplisten john\n\n"
;
    }

    void stoppedCb(exception const& e) override {
        myCout_ << (e.what()) << endl;
    };

    /**
     * @brief callback for JustBytes
     * @details it is the user's responsibility to figure ouot the size of the message
     * thru tag and interpret the bytes
     * 
     * @param tag message tag - starting from 1000
     * @param bytes contents of the message - will not exceed the associated 
     * Context::maxMessageSize() return value
     */
    void handleJustBytesCb(uint16_t tag, uint8_t* bytes) {
        myCout_ << dec << tag;
        hasMemoryAttachment* att = nullptr;
        if (memoryAttachmentTags_.find(tag) != memoryAttachmentTags_.end()) {
            att = reinterpret_cast<hasMemoryAttachment*>(bytes);
            bytes += sizeof(hasMemoryAttachment);
        }
        if (hex_) {
            for (auto p = bytes; p != bytes + ctx_.maxMessageSize(); ++p) {
                myCout_ << ' ' << boost::format("%02x") % (uint16_t)*p;
            }
            if (att) {
                myCout_ << '\n';
                for (auto p = (uint8_t*)att->attachment; p != (uint8_t*)att->attachment + att->len; ++p) {
                    myCout_ << boost::format("%02x") % (uint16_t)*p << ' ';
                }
            }
        } else {
            myCout_ << ' ' << string((char*)bytes, strnlen((char*)bytes, ctx_.maxMessageSize()));
            if (att) {
                myCout_ << '\n';
                myCout_ << string((char*)att->attachment, att->len);
            }
        }
        myCout_ << endl;
    }
    
    /**
     * @brief wait until the console closes
     */
    void waitUntilFinish() {
        stdinThread_.join();
    }

 private:
    void stdinThreadEntrance() {
        processCmd(initCmd_);
        processCmd(myCin_);
    }

    void processCmd(istream& is) {
        string line; 
        while(getline(is, line)) {
            if (!line.size()) continue;
            istringstream iss{line};
            string op;
            iss >> op;
            if (op == "topic") {
                string topic;
                iss >> topic;
                Topic tmp(topic.c_str());
                auto sender = NetCtx::instance().getSender(tmp);
                if (sender) {
                    topic_ = tmp;
                    sender_ = sender;
                } else {
                    myCerr_ << " cannot find transport for " << topic << endl;
                }
            } else if (op == "sendstr") {
                if (!sender_) {
                    myCerr_ << " topic is not set yet" << endl;
                    continue;
                }
                uint16_t tag;
                iss >> tag;
                string msg;
                getline(iss, msg);
                if (iss) {
                    sender_->sendBytes(tag, msg.c_str() + 1, //drop leading space
                        min(msg.size(), ctx_.maxMessageSize())
                    );
                } else {
                    myCerr_ << " sendstr syntax error" << endl;
                }
            } else if (op == "sendstratt") {
                if (!sender_) {
                    myCerr_ << " topic is not set yet" << endl;
                    continue;
                }
                uint16_t tag;
                iss >> tag;
                string msg;
                getline(iss, msg);
                if (!iss || !getline(is, line)) {
                    myCerr_ << " sendstratt syntax error" << endl;
                    continue;
                }
                auto att = malloc(line.size());
                memcpy(att, line.c_str(), line.size());
                auto rawSize = sizeof(hasMemoryAttachment) + msg.size();
                char rawMsg[rawSize];
                auto hma = new (rawMsg) hasMemoryAttachment;
                hma->attachment = att;
                hma->len = line.size();
                memcpy((char*)rawMsg + sizeof(hasMemoryAttachment), msg.c_str() + 1, msg.size()); //drop leading space
                sender_->sendBytes(tag, hma, min(rawSize, ctx_.maxMessageSize()));
            } else if (op == "sendhex") {
                if (!sender_) {
                    myCerr_ << " topic is not set yet" << endl;
                    continue;
                }
                uint16_t tag;
                iss >> tag;
                iss >> hex;
                vector<uint8_t> bytes;
                uint16_t tmp;
                while (iss >> tmp) {
                    bytes.push_back((uint8_t)tmp);
                }

                if (iss.eof()) {
                    sender_->sendBytes(tag, &bytes[0], 
                        min(bytes.size(), ctx_.maxMessageSize())
                    );
                } else {
                    myCerr_ << " sendhex syntax error" << endl;
                }
            } else if (op == "sendhexatt") {
                if (!sender_) {
                    myCerr_ << " topic is not set yet" << endl;
                    continue;
                }
                
                uint16_t tag;
                iss >> tag;
                iss >> hex;
                vector<uint8_t> bytes;
                uint16_t tmp;
                while (iss >> tmp) {
                    bytes.push_back((uint8_t)tmp);
                }
                if (!iss.eof() || !getline(is, line)) {
                    myCerr_ << " sendhexhex syntax error" << endl;
                    continue;
                }

                istringstream iss{line};
                iss >> hex;
                vector<uint8_t> attBytes;
                while (iss >> tmp) {
                    attBytes.push_back((uint8_t)tmp);
                }

                auto att = malloc(attBytes.size());
                memcpy(att, &(attBytes[0]), attBytes.size());
                auto rawSize = sizeof(hasMemoryAttachment) + attBytes.size();
                char rawMsg[rawSize];
                auto hma = new (rawMsg) hasMemoryAttachment;
                hma->attachment = att;
                hma->len = attBytes.size();
                memcpy((char*)rawMsg + sizeof(hasMemoryAttachment), &bytes[0], bytes.size());
                sender_->sendBytes(tag, hma, min(rawSize, ctx_.maxMessageSize()));
            } else if (op == "ohex") {
                hex_ = true;
            } else if (op == "ostr") {
                hex_ = false;
            } else if (op == "listen") {
                string topic;
                iss >> topic;
                NetCtx::instance().listenTo(Topic(topic));
            } else if (op == "stoplisten") {
                string topic;
                iss >> topic;
                NetCtx::instance().stopListenTo(Topic(topic));
            } else {
                myCerr_ << " unknown command " << op << endl;
            }
        }
    }

    hmbdc::app::Context<>& ctx_;
    atomic<bool> hex_;
    thread stdinThread_;
    istream& myCin_;
    ostream& myCout_;
    ostream& myCerr_;
    Topic topic_;
    typename NetCtx::Sender* sender_;
    istringstream initCmd_;
    std::unordered_set<uint16_t> memoryAttachmentTags_;
};

} //consoleclient_detail

template <typename NetCtx>
using ConsoleClient = consoleclient_detail::ConsoleClient<NetCtx>;
}}}

